#!/usr/bin/env python2.2

"""
Execution framework for SML/NG optimizer.  Takes care of
  * timeouts (cancelling optimization after a set period)
  * saving the original input in case we have to give up (timeout,
    or "optimized" output ends up bigger than input)

"""

import sys, os
import time
import tempfile
import traceback

from sml_parser import parse
from list_transforms import collapse_whitespace, merge_chunks, \
     merge_adjacent_whitespace
from tree import build_trees
from tree_transforms import merge_simple, merge_color, merge_size, \
     lift_simple_tag, lift_color, lift_size


# Time-consuming code is responsible for periodically checking this,
# and getting the hell out if ABORT_TIME and time.time() >= ABORT_TIME
ABORT_TIME = sys.maxint


class TimeoutError (Exception):
    pass

def checktime ():
    if time.time() >= (ABORT_TIME - 10):
        raise TimeoutError

def writenow (msg, file=None):
    if file is None:
        file = sys.stderr
    file.write(msg)
    file.flush()


def copy_data (in_file, out_file):
    size = 0
    while 1:
        data = in_file.read(8192)
        #print "got %d bytes" % len(data)
        if not data: break
        size += len(data)
        out_file.write(data)
    return size


def spew_saved_input (filename):
    file = open(filename)
    copy_data(file, sys.stdout)
    file.close()


def optimize (tree):
    while 1:
        simple_merges = 0
        for tag in ("B", "I", "TT", "U", "EM", "S"):
            lift_simple_tag(tree, tag)
            simple_merges += merge_simple(tree)
            #writenow("after lift_simple_tag(..., %s):\n" % tag); tree.dump()
            checktime()
        if simple_merges == 0:
            break
        writenow("simple_merge removed %d node(s).\n" % simple_merges)

    while 1:
        lift_color(tree)
        color_merges = merge_color(tree)
        #writenow("after merge_color():\n"); tree.dump()
        if color_merges == 0:
            break
        writenow("color_merge removed %d node(s).\n" % color_merges)
        checktime()

    while 1:
        lift_size(tree)
        size_merges = merge_size(tree)
        #writenow("after merge_size():\n"); tree.dump()
        if size_merges == 0:
            break
        writenow("size_merge removed %d node(s).\n" % size_merges)
        checktime()


def main ():
    global ABORT_TIME

    args = sys.argv[1:]
    try:
        time_limit = int(args[0])
    except (IndexError, ValueError):
        sys.stderr.write("warning: no time limit given\n")
        time_limit = None

    start_time = time.time()
    if time_limit:
        ABORT_TIME = start_time + time_limit


    # Save the original stdin to a temporary file.  That way, if we have
    # to give up on optimization (run out of time, ends up being
    # bigger), we can fallback on something that's no worse than the
    # original input.
    tmp_filename = tempfile.mktemp(".sml")
    tmp_file = open(tmp_filename, "w")
    #print "copying stdin to %s" % tmp_file
    input_size = copy_data(sys.stdin, tmp_file)
    tmp_file.close()

    try:
        try:
            writenow("initiating parse and list transformations...\n")
            in_file = open(tmp_filename)
            chunks = parse(in_file)
            checktime()
            chunks = merge_chunks(chunks)
            checktime()
            chunks = collapse_whitespace(chunks)
            checktime()
            chunks = merge_adjacent_whitespace(chunks)
            checktime()

            #print "optimized chunks:"
            #writenow("forcing completion...\n")
            #for c in chunks: print c

            output = ""
            writenow("building trees...\n")
            for tree in build_trees(chunks):
                #writenow("initial tree:\n"); tree.dump()

                writenow("performing tree transformations...\n")
                optimize(tree)
                #writenow("final tree:\n"); tree.dump()
                
                writenow("rendering...\n")
                output += "".join(list(tree.render()))

            n = len(output)
            if n < input_size:
                writenow("removed %s bytes (%.2f)\n" % 
                         (input_size - n, float(n)/input_size))
                # we did some good
                sys.stdout.write(output)
                #sys.stdout.write("\n") # end with a newline
            else:
                writenow("didn't save any bytes, spewing original\n")
                spew_saved_input(tmp_filename)
                

        except TimeoutError:
            sys.stderr.write("timed out after %d sec\n" %
                             (time.time() - start_time))
            spew_saved_input(tmp_filename)

        except KeyboardInterrupt:
            sys.exit("interrupted")

        except MemoryError:
            sys.stderr.write("ran out of memory\n")
            spew_saved_input(tmp_filename)

        except:
            (exc_type, exc_val, tb) = sys.exc_info()
            sys.stderr.write("caught %s: %s\n" % (exc_type, exc_val))
            spew_saved_input(tmp_filename)
    finally:
        try:
            os.remove(tmp_filename)
        except OSError:
            pass

# main ()

main()
